@@ -9,10 +9,13 @@ import shared, js_optimizer
 from tempfiles import try_delete
 
 js_file = sys.argv[1]
-mem_init_file = sys.argv[2]
+binary_file = sys.argv[2] # mem init for js, wasm binary for wasm
 total_memory = int(sys.argv[3])
 total_stack = int(sys.argv[4])
 global_base = int(sys.argv[5])
+binaryen_bin = sys.argv[6]
+
+wasm = not not binaryen_bin
 
 assert global_base > 0
 
@@ -32,7 +35,17 @@ def find_ctors(js):
   ctors_end += 3
   return (ctors_start, ctors_end)
 
-def eval_ctors(js, mem_init, num):
+def find_ctors_data(js, num):
+  ctors_start, ctors_end = find_ctors(js)
+  assert ctors_start > 0
+  ctors_text = js[ctors_start:ctors_end]
+  all_ctors = filter(lambda ctor: ctor.endswith('()') and not ctor == 'function()' and '.' not in ctor, ctors_text.split(' '))
+  all_ctors = map(lambda ctor: ctor.replace('()', ''), all_ctors)
+  assert len(all_ctors) > 0
+  ctors = all_ctors[:num]
+  return ctors_start, ctors_end, all_ctors, ctors
+
+def eval_ctors_js(js, mem_init, num):
 
   def kill_func(asm, name):
     before = len(asm)
@@ -48,14 +61,7 @@ def eval_ctors(js, mem_init, num):
     return asm
 
   # Find the global ctors
-  ctors_start, ctors_end = find_ctors(js)
-  assert ctors_start > 0
-  ctors_text = js[ctors_start:ctors_end]
-  all_ctors = filter(lambda ctor: ctor.endswith('()') and not ctor == 'function()' and '.' not in ctor, ctors_text.split(' '))
-  all_ctors = map(lambda ctor: ctor.replace('()', ''), all_ctors)
-  total_ctors = len(all_ctors)
-  assert total_ctors > 0
-  ctors = all_ctors[:num]
+  ctors_start, ctors_end, all_ctors, ctors = find_ctors_data(js, num)
   shared.logging.debug('trying to eval ctors: ' + ', '.join(ctors))
   # Find the asm module, and receive the mem init.
   asm = get_asm(js)
@@ -257,6 +263,7 @@ console.log(JSON.stringify([numSuccessful, Array.prototype.slice.call(heap.subar
   # out contains the new mem init and other info
   num_successful, mem_init_raw, atexits = json.loads(out_result)
   mem_init = ''.join(map(chr, mem_init_raw))
+  total_ctors = len(all_ctors)
   if num_successful < total_ctors:
     shared.logging.debug('not all ctors could be evalled, something was used that was not safe (and therefore was not defined, and caused an error):\n========\n' + err_result + '========')
   # Remove the evalled ctors, add a new one for atexits if needed, and write that out
@@ -272,6 +279,23 @@ console.log(JSON.stringify([numSuccessful, Array.prototype.slice.call(heap.subar
   js = js[:ctors_start] + new_ctors + js[ctors_end:]
   return (num_successful, js, mem_init, ctors)
 
+def eval_ctors_wasm(js, wasm_file, num):
+  ctors_start, ctors_end, all_ctors, ctors = find_ctors_data(js, num)
+  cmd = [os.path.join(binaryen_bin, 'wasm-ctor-eval'), wasm_file, '-o', wasm_file, '--ctors=' + ','.join(ctors)]
+  shared.logging.debug('wasm ctor cmd: ' + str(cmd))
+  out, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
+  num_successful = err.count('success on')
+  shared.logging.debug(err)
+  if len(ctors) == len(all_ctors):
+    new_ctors = ''
+  else:
+    elements = []
+    for ctor in all_ctors[num:]:
+      elements.append('{ func: function() { %s() } }' % ctor)
+    new_ctors = '__ATINIT__.push(' + ', '.join(elements) + ');'
+  js = js[:ctors_start] + new_ctors + js[ctors_end:]
+  return num_successful, js
+
 # main
 if __name__ == '__main__':
   js = open(js_file).read()
@@ -288,50 +312,63 @@ if __name__ == '__main__':
   num_ctors = ctors_text.count('function()')
   shared.logging.debug('ctor_evaller: %d ctors, from |%s|' % (num_ctors, ctors_text))
 
-  if os.path.exists(mem_init_file):
-    mem_init = json.dumps(map(ord, open(mem_init_file, 'rb').read()))
+  if not wasm:
+    # js path
+    mem_init_file = binary_file
+    if os.path.exists(mem_init_file):
+      mem_init = json.dumps(map(ord, open(mem_init_file, 'rb').read()))
+    else:
+      mem_init = []
+
+    # find how many ctors we can remove, by bisection (if there are hundreds, running them sequentially is silly slow)
+
+    shared.logging.debug('ctor_evaller: trying to eval %d global constructors' % num_ctors)
+    num_successful, new_js, new_mem_init, removed = eval_ctors_js(js, mem_init, num_ctors)
+    if num_successful == 0:
+      shared.logging.debug('ctor_evaller: not successful')
+      sys.exit(0)
+
+    shared.logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
+    if num_successful == num_ctors:
+      js = new_js
+      mem_init = new_mem_init
+    else:
+      shared.logging.debug('ctor_evaller: final execution')
+      check, js, mem_init, removed = eval_ctors_js(js, mem_init, num_successful)
+      assert check == num_successful
+    open(js_file, 'w').write(js)
+    open(mem_init_file, 'wb').write(mem_init)
+
+    # Dead function elimination can help us
+
+    shared.logging.debug('ctor_evaller: eliminate no longer needed functions after ctor elimination')
+    # find exports
+    asm = get_asm(open(js_file).read())
+    exports_start = asm.find('return {')
+    exports_end = asm.find('};', exports_start)
+    exports_text = asm[asm.find('{', exports_start) + 1 : exports_end]
+    exports = map(lambda x: x.split(':')[1].strip(), exports_text.replace(' ', '').split(','))
+    for r in removed:
+      assert r in exports, 'global ctors were exported'
+    exports = filter(lambda e: e not in removed, exports)
+    # fix up the exports
+    js = open(js_file).read()
+    absolute_exports_start = js.find(exports_text)
+    js = js[:absolute_exports_start] + ', '.join(map(lambda e: e + ': ' + e, exports)) + js[absolute_exports_start + len(exports_text):]
+    open(js_file, 'w').write(js)
+    # find unreachable methods and remove them
+    reachable = shared.Building.calculate_reachable_functions(js_file, exports, can_reach=False)['reachable']
+    for r in removed:
+      assert r not in reachable, 'removed ctors must NOT be reachable'
+    shared.Building.js_optimizer(js_file, ['removeFuncs'], extra_info={ 'keep': reachable }, output_filename=js_file)
   else:
-    mem_init = []
-
-  # find how many ctors we can remove, by bisection (if there are hundreds, running them sequentially is silly slow)
-
-  shared.logging.debug('ctor_evaller: trying to eval %d global constructors' % num_ctors)
-  num_successful, new_js, new_mem_init, removed = eval_ctors(js, mem_init, num_ctors)
-  if num_successful == 0:
-    shared.logging.debug('ctor_evaller: not successful')
-    sys.exit(0)
-
-  shared.logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
-  if num_successful == num_ctors:
-    js = new_js
-    mem_init = new_mem_init
-  else:
-    shared.logging.debug('ctor_evaller: final execution')
-    check, js, mem_init, removed = eval_ctors(js, mem_init, num_successful)
-    assert check == num_successful
-  open(js_file, 'w').write(js)
-  open(mem_init_file, 'wb').write(mem_init)
-
-  # Dead function elimination can help us
-
-  shared.logging.debug('ctor_evaller: eliminate no longer needed functions after ctor elimination')
-  # find exports
-  asm = get_asm(open(js_file).read())
-  exports_start = asm.find('return {')
-  exports_end = asm.find('};', exports_start)
-  exports_text = asm[asm.find('{', exports_start) + 1 : exports_end]
-  exports = map(lambda x: x.split(':')[1].strip(), exports_text.replace(' ', '').split(','))
-  for r in removed:
-    assert r in exports, 'global ctors were exported'
-  exports = filter(lambda e: e not in removed, exports)
-  # fix up the exports
-  js = open(js_file).read()
-  absolute_exports_start = js.find(exports_text)
-  js = js[:absolute_exports_start] + ', '.join(map(lambda e: e + ': ' + e, exports)) + js[absolute_exports_start + len(exports_text):]
-  open(js_file, 'w').write(js)
-  # find unreachable methods and remove them
-  reachable = shared.Building.calculate_reachable_functions(js_file, exports, can_reach=False)['reachable']
-  for r in removed:
-    assert r not in reachable, 'removed ctors must NOT be reachable'
-  shared.Building.js_optimizer(js_file, ['removeFuncs'], extra_info={ 'keep': reachable }, output_filename=js_file)
+    # wasm path
+    wasm_file = binary_file
+    shared.logging.debug('ctor_evaller (wasm): trying to eval %d global constructors' % num_ctors)
+    num_successful, new_js = eval_ctors_wasm(js, wasm_file, num_ctors)
+    if num_successful == 0:
+      shared.logging.debug('ctor_evaller: not successful')
+      sys.exit(0)
+    shared.logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
+    open(js_file, 'w').write(new_js)
 
